package managed_resources //nolint:stylecheck

import (
	"context"
	"fmt"

	"k8s.io/apimachinery/pkg/api/meta"
	metav1 "k8s.io/apimachinery/pkg/apis/meta/v1"
	"k8s.io/apimachinery/pkg/apis/meta/v1/unstructured"
	"k8s.io/apimachinery/pkg/runtime/schema"
	"k8s.io/client-go/dynamic"
)

type Object struct {
	GVK       schema.GroupVersionKind
	Namespace string
	Name      string
}

type ObjectError struct {
	Object
	Error error
}

type ObjectManager struct {
	Client       dynamic.Interface
	Mapper       meta.RESTMapper
	FieldManager string
}

func (m *ObjectManager) Apply(ctx context.Context, objs []*unstructured.Unstructured) ([]*unstructured.Unstructured, []ObjectError) {
	clusterScoped, namespaced := splitNamespacedAndClusterScoped(objs)

	// Must dry run and then apply Namespace, etc so that namespaced resources pass their dry run (namespace has to exist)
	clusterScoped, errs := m.dryApply(ctx, clusterScoped)
	if len(errs) > 0 {
		return clusterScoped, errs
	}

	namespaced, errs = m.dryApply(ctx, namespaced)
	if len(errs) > 0 {
		return append(clusterScoped, namespaced...), errs
	}

	return append(clusterScoped, namespaced...), nil
}

func (m *ObjectManager) dryApply(ctx context.Context, objs []*unstructured.Unstructured) ([]*unstructured.Unstructured, []ObjectError) {
	_, errs := m.apply(ctx, objs, true)
	if len(errs) > 0 {
		return nil, errs // do not return anything from a dry run
	}

	return m.apply(ctx, objs, false)
}

func (m *ObjectManager) apply(ctx context.Context, objs []*unstructured.Unstructured, dryRun bool) ([]*unstructured.Unstructured, []ObjectError) {
	var dryRunObjs []string

	if dryRun {
		dryRunObjs = []string{metav1.DryRunAll}
	}

	applied := make(chan *unstructured.Unstructured)
	errs := make(chan ObjectError)

	for _, obj := range objs {
		go func() {
			gvk := obj.GetObjectKind().GroupVersionKind()
			ns := obj.GetNamespace()
			name := obj.GetName()
			mapping, err := m.Mapper.RESTMapping(gvk.GroupKind(), gvk.Version)
			if err != nil {
				errs <- ObjectError{
					Object: Object{
						GVK:       gvk,
						Namespace: ns,
						Name:      name,
					},
					Error: fmt.Errorf("mapper: %w", err),
				}
				return
			}
			resp, err := m.Client.
				Resource(mapping.Resource).
				Namespace(ns). // it's ok if this is an empty string
				Apply(ctx, name, obj, metav1.ApplyOptions{
					Force:        true,
					FieldManager: m.FieldManager,
					DryRun:       dryRunObjs,
				})
			if err != nil {
				errs <- ObjectError{
					Object: Object{
						GVK:       gvk,
						Namespace: ns,
						Name:      name,
					},
					Error: fmt.Errorf("apply: %w", err),
				}
				return
			}
			applied <- resp
		}()
	}

	var retErrs []ObjectError
	retObjs := make([]*unstructured.Unstructured, 0, len(objs))
	for len(retObjs)+len(retErrs) < len(objs) {
		select {
		case obj := <-applied:
			retObjs = append(retObjs, obj)
		case e := <-errs:
			if dryRun {
				e.Error = fmt.Errorf("dry run: %w", e.Error)
			}
			retErrs = append(retErrs, e)
		}
	}
	return retObjs, retErrs
}

func splitNamespacedAndClusterScoped(objs []*unstructured.Unstructured) ([]*unstructured.Unstructured, []*unstructured.Unstructured) {
	var clusterScoped []*unstructured.Unstructured
	var namespaced []*unstructured.Unstructured
	for _, obj := range objs {
		if obj.GetNamespace() == "" {
			clusterScoped = append(clusterScoped, obj)
		} else {
			namespaced = append(namespaced, obj)
		}
	}
	return clusterScoped, namespaced
}
