package managed_resources //nolint:stylecheck

import (
	"bytes"
	"encoding/json"
	"fmt"
	"text/template"

	"gitlab.com/gitlab-org/cluster-integration/gitlab-agent/v17/internal/api"
	"k8s.io/apimachinery/pkg/api/meta"
	"k8s.io/apimachinery/pkg/apis/meta/v1/unstructured"
	"sigs.k8s.io/yaml"
)

type Renderer struct {
	Mapper meta.RESTMapper
}

// RenderEnvironmentTemplate renders a template.
// All returned errors are an instance of RendererError.
func (r *Renderer) RenderEnvironmentTemplate(tpl *EnvironmentTemplate, info *TemplatingInfo,
	processRenderedObj func(*unstructured.Unstructured) error) (*RenderedEnvironmentTemplate, error) {

	// 1. Unmarshal
	var raw EnvironmentTemplateData[json.RawMessage]
	err := yaml.UnmarshalStrict(tpl.Data, &raw)
	if err != nil {
		return nil, NewErrorf(InvalidArgument, "Invalid template: %v", err)
	}

	// 2. Render - templating
	vars := info2vars(info)
	renderedObjs := make([]unstructured.Unstructured, 0, len(raw.Objects))
	for i, o := range raw.Objects {
		renderedObject, rendErr := renderObject(o, vars)
		if rendErr != nil {
			return nil, NewErrorf(InvalidArgument, "Invalid template in object[%d]: %v", i, rendErr)
		}
		renderedObjs = append(renderedObjs, renderedObject)
	}

	// 3. Render - process objects
	err = r.processObjects(renderedObjs, tpl, info, processRenderedObj)
	if err != nil {
		return nil, NewErrorf(InvalidArgument, "Invalid template: processing objects: %v", err)
	}

	// 4. Return result
	return renderedObjs2renderedTemplate(tpl.Name, renderedObjs, raw.EnvironmentTemplateDataFields)
}

func renderedObjs2renderedTemplate(name string, renderedObjs []unstructured.Unstructured, fields EnvironmentTemplateDataFields) (*RenderedEnvironmentTemplate, error) {
	outObjs := make([]map[string]any, 0, len(renderedObjs))
	for _, obj := range renderedObjs {
		outObjs = append(outObjs, obj.Object)
	}
	data, err := yaml.Marshal(&EnvironmentTemplateData[map[string]any]{
		Objects:                       outObjs,
		EnvironmentTemplateDataFields: fields,
	})
	if err != nil {
		return nil, NewErrorf(Internal, "YAML marshal: %v", err)
	}
	return &RenderedEnvironmentTemplate{
		Name: name,
		Data: data,
	}, nil
}

func (r *Renderer) processObjects(objs []unstructured.Unstructured, tpl *EnvironmentTemplate, info *TemplatingInfo,
	processRenderedObj func(*unstructured.Unstructured) error) error {

	// 1. Prepare labels and annotations
	envInfoAnnValBytes, err := json.Marshal(info.ToEnvironmentInfo(tpl.Name))
	if err != nil {
		return err
	}
	envInfoAnnVal := string(envInfoAnnValBytes)

	envInfoKey := fmt.Sprintf("%s/env-%s-%d-%d",
		api.AgentKeyPrefix,
		info.Environment.Slug,
		info.Project.Id,
		info.Agent.Id,
	)
	labelKeys := []string{
		envInfoKey,
		fmt.Sprintf("%s-%d", api.AgentIDKey, info.Agent.Id),
		fmt.Sprintf("%s-%d", api.ProjectIDKey, info.Project.Id),
		fmt.Sprintf("%s-%s", api.EnvironmentSlugKey, info.Environment.Slug),
	}

	for i := range objs {
		obj := &objs[i]

		// 2. Set environment info annotation and various labels
		setEnvInfoAnnotation(obj, envInfoKey, envInfoAnnVal)
		setEnvInfoLabels(obj, labelKeys)

		// 3. Call caller-provided processor func
		err = processRenderedObj(obj)
		if err != nil {
			return err
		}
	}
	return nil
}

func setEnvInfoAnnotation(obj *unstructured.Unstructured, envInfoKey, envInfoAnnVal string) {
	a := obj.GetAnnotations()
	if a == nil {
		a = map[string]string{}
	}
	a[envInfoKey] = envInfoAnnVal
	obj.SetAnnotations(a)
}

func setEnvInfoLabels(obj *unstructured.Unstructured, labelKeys []string) {
	l := obj.GetLabels()
	if l == nil {
		l = map[string]string{}
	}
	for _, k := range labelKeys {
		l[k] = ""
	}
	obj.SetLabels(l)
}

func info2vars(info *TemplatingInfo) map[string]any {
	return map[string]any{
		"agent": map[string]any{
			"id":   info.Agent.Id,
			"name": info.Agent.Name,
			"url":  info.Agent.Url,
		},
		"environment": map[string]any{
			"id":       info.Environment.Id,
			"name":     info.Environment.Name,
			"slug":     info.Environment.Slug,
			"url":      info.Environment.GetUrl(), // empty string if not provided
			"page_url": info.Environment.PageUrl,
			"tier":     info.Environment.Tier,
		},
		"project": map[string]any{
			"id":   info.Project.Id,
			"slug": info.Project.Slug,
			"path": info.Project.Path,
			"url":  info.Project.Url,
		},
		"ci_pipeline": map[string]any{
			"id": info.Pipeline.Id,
		},
		"ci_job": map[string]any{
			"id": info.Job.Id,
		},
		"user": map[string]any{
			"id":       info.User.Id,
			"username": info.User.Username,
		},
	}
}

func renderObject(src []byte, vars map[string]any) (unstructured.Unstructured, error) {
	var buf bytes.Buffer
	t := template.New("object").Option("missingkey=error")
	t, err := t.Parse(string(src))
	if err != nil {
		return unstructured.Unstructured{}, fmt.Errorf("parse: %w", err)
	}
	err = t.Execute(&buf, vars)
	if err != nil {
		return unstructured.Unstructured{}, fmt.Errorf("execute: %w", err)
	}
	var renderedObject map[string]any
	err = yaml.UnmarshalStrict(buf.Bytes(), &renderedObject)
	if err != nil {
		return unstructured.Unstructured{}, fmt.Errorf("resulting YAML is invalid: %w", err)
	}
	return unstructured.Unstructured{Object: renderedObject}, nil
}

type EnvironmentTemplateData[O any] struct {
	Objects []O `json:"objects"`
	EnvironmentTemplateDataFields
}

// EnvironmentTemplateDataFields holds all fields except objects.
// This simplifies copying of these fields.
type EnvironmentTemplateDataFields struct {
	ApplyResources  string `json:"apply_resources"`
	DeleteResources string `json:"delete_resources"`
}
